#! /usr/bin/env ruby
=begin

General-purpose digital circuit simulator Andap

memo
token format
[ [CLASS_TYPE, :ID] [TYPE, "????"] ...... ]

[ [Element,  :ID] | [:IDENT, "ID"]
[ [Delay,    :ID] | [:FLORT, "2.0"]
__________________| [:DLY,   "ID"]
__________________| ["+",    "+"]

=end

require 'bigdecimal'

def usage
  $\ = "\n"
  STDERR.print "Usage : #{File.basename(__FILE__)} <circuit-file> [input-file] [output-file]"
  exit
end

module Andap

  class AndapError < StandardError
    CircuitFormatError   = "Error in circuit format".freeze
    OutputNotDefineError = "OUTPUT is not defined".freeze
    InternalBugError     = "Maybe this error is Andap's bug!".freeze
  end

  ## Symbol => Element Object
  Circuit = {}

  class ::Hash
    def []( key )
      self.fetch( key ) {|key|
        raise AndapError, "#{key.to_s} is not defined"
      }
    end
  end

  class Element
    def initialize( element_name )
      if ( Circuit.has_key?( element_name ) )
        raise AndapError, "Element #{element_name} is already defined"
      end
      Circuit[element_name] = self

      @val = BigDecimal( "0.0" )
      @already = true
    end

    def val
      unless @already
        @already = true if self.instance_of?( Delay )
        calculate_value
        @already = true
      end
      return @val
    end

    def reset
      @already = false
    end
  end


  class Delay < Element
    def initialize( element_name )
      super
      @delayed_value = BigDecimal( "0.0" )
    end

    def set_forward( sym )
      @val = @delayed_value
      @delayed_value = Circuit[sym].val
    end
  end

  Element.new( :INPUT )
  class << Circuit[:INPUT]
    def val=( val )
      @val = val
      Circuit.each_value{|one| one.reset }
    end

    def calculate_value
      @val
    end
  end


  def self.scan( file )
    IO.foreach( file ) {|line|
      line = line.strip.sub( /\#.*/, '' )
      next if line.empty?

      unless /^\w+/ =~ line
        raise AndapError, AndapError::CircuitFormatError
      end
      tokens = [ [Element, $&.intern] ]
      line = $'.strip

      unless /^=/ =~ line
        raise AndapError, AndapError::CircuitFormatError
      end
      line = $'.strip

      until line.empty?
        case line
        when /^DLY\(\s*(\w+)\s*\)/
          unless ( tokens.size == 1 )
            raise AndapError, AndapError::CircuitFormatError
          end
          tokens[0][0] = Delay
          tokens << [:DLY, $1]

        when /^[0-9]+\.[0-9]+/
          tokens << [:FLORT, $&]

        when /^\w+/
          tokens << [:IDENT, $&]

        when /^[^\s]+/
          tokens << [$&, $&]

        else
          raise AndapError, AndapError::InternalBugError
        end
        line = $'.strip
      end
      make_elements tokens
    }
    if ( Circuit[:OUTPUT].nil? )
      raise AndapError, AndapError::OutputNotDefineError
    end
  end

  def self.make_elements( token )
    self.module_eval( <<-EOS
      ele = #{token[0][0].name}.new( %s!#{token[0][1]}! )
      class << ele
        def calculate_value
          #{ str = ""
            if ( token[1][0] == %s!DLY! )
              str = "set_forward( %s!#{token[1][1]}! )"
            else
              str << "@val = "
              token[1..-1].each{|one|
                case ( one[0] )
                when :IDENT
                  str << "Circuit[%s!#{one[1]}!].val "
                when :FLORT
                  str << "BigDecimal( #{one[1]}.to_s ) "
                else
                  str << "#{one[1]} "
                end
              }
            end
            str
          }
        end
      end
      EOS
    )
  end

  def self.one_time( input )
    Circuit[:INPUT].val = BigDecimal( input.to_s )
    Circuit[:OUTPUT].val.to_f
  end

end

## Main
usage if ( ARGV.size < 1 || ARGV.size > 3 )

Andap::scan( ARGV[0] )

if ARGV[1].nil?
  print Andap::one_time( 1.0 ), "\n"
  44099.times{ printf( "%+.17f\n", Andap::one_time( 0.0 ) ) }
  exit
end

input = File.open( ARGV[1], "r" )
output = if ( ARGV[2].nil? ) then STDOUT else File.open( ARGV[2], "w" ) end

input.each{|line|
  line = line.strip.sub( /\#.*/, '' )
  next if line.empty?
  output.printf( "%+.17f\n", Andap::one_time( line.to_f ) )
}

input.close
output.close

